"""
Contains utility functions for mity modules.
"""
import logging
import os
import subprocess
import sys
from glob import glob
import pysam


class MityUtil:
    """
    Contains utility functions for mity modules.
    """

    MITY_DIR = "mitylib"
    GENOME_FILE = "mitylib/reference/b37d5.genome"
    REF_DIR = "reference"
    ANNOT_DIR = "annot"

    @classmethod
    def get_mity_dir(cls):
        """
        Get the directory path of the Mity library.

        Returns:
            str: The path to the Mity library directory.
        """
        path = os.path.dirname(sys.modules["mitylib"].__file__)
        return path

    @classmethod
    def tabix(cls, f: str):
        """
        Generate a tabix index for a bgzipped file.

        Parameters:
            f (str): The path to a bgzip compressed file.

        Returns:
            None
        """
        tabix_call = "tabix -f " + f
        logging.debug(tabix_call)
        subprocess.run(tabix_call, shell=True, check=False)

    @classmethod
    def select_reference_fasta(cls, reference: str, custom_reference_fa: str = None):
        """
        Select the reference genome fasta file.

        Parameters:
            reference (str): One of the inbuilt reference genomes: hs37d5, hg19, hg38, mm10.
            custom_reference_fa (str, optional): The path to a custom reference genome, or None.

        Returns:
            str: The path to the selected reference genome fasta file.
        """
        if custom_reference_fa is not None and os.path.exists(custom_reference_fa):
            res = custom_reference_fa
        else:
            ref_dir = os.path.join(cls.get_mity_dir(), cls.REF_DIR)
            res = glob(f"{ref_dir}/{reference}.*.fa")
            logging.debug(",".join(res))
            assert len(res) == 1
            res = res[0]
        return res

    @classmethod
    def select_reference_genome(
        cls, reference: str, custom_reference_genome: str = None
    ):
        """
        Select the reference genome .genome file.

        Parameters:
            reference (str): One of the inbuilt reference genomes: hs37d5, hg19, hg38, mm10.
            custom_reference_genome (str, optional): The path to a custom reference .genome file, or None.

        Returns:
            str: The path to the selected reference .genome file.
        """
        if custom_reference_genome is not None and os.path.exists(
            custom_reference_genome
        ):
            res = custom_reference_genome
        else:
            ref_dir = os.path.join(cls.get_mity_dir(), cls.REF_DIR)
            logging.debug("Looking for .genome file in %s", ref_dir)
            res = glob(f"{ref_dir}/{reference}.genome")
            logging.debug(",".join(res))
            assert len(res) == 1
            res = res[0]
        return res

    @classmethod
    def vcf_get_mt_contig(cls, vcf: str):
        """
        Get the mitochondrial contig name and length from a VCF file.

        Parameters:
            vcf (str): Path to a VCF file.

        Returns:
            tuple: A tuple of contig name as a str and length as an int.
        """
        r = pysam.VariantFile(vcf, "r")
        chroms = r.header.contigs
        mito_contig = set(["MT", "chrM"]).intersection(chroms)
        assert len(mito_contig) == 1
        mito_contig = "".join(mito_contig)
        return r.header.contigs[mito_contig].name, r.header.contigs[mito_contig].length

    @classmethod
    def get_annot_file(cls, annotation_file_path: str):
        """
        Get the path to an annotation file.

        Parameters:
            annotation_file_path (str): The name of the annotation file.

        Returns:
            str: The path to the annotation file.
        """
        mitylibdir = cls.get_mity_dir()
        path = os.path.join(mitylibdir, cls.ANNOT_DIR, annotation_file_path)
        assert os.path.exists(path)
        return path

    @classmethod
    def make_prefix(cls, vcf_path: str):
        """
        Make a prefix based on the input vcf path. This handles vcf files from
        previous steps of mity. e.g. from call to normalise, etc.

        Format of MITY output filenames:
            prefix.mity.call.vcf.gz
            prefix.mity.normalise.vcf.gz
            prefix.mity.merge.vcf.gz
            prefix.report.xlsx
        """

        prefix = (
            os.path.basename(vcf_path)
            .replace(".mity", "")
            .replace(".call", "")
            .replace(".normalise", "")
            .replace(".merge", "")
            .replace(".report", "")
            .replace(".vcf.gz", "")
        )

        return prefix

    @classmethod
    def gsort(cls, input_path: str, output_path: str, genome: str):
        """
        Run gsort.
        """
        gsort_cmd = f"gsort {input_path} {genome} | bgzip -cf > {output_path}"
        subprocess.run(gsort_cmd, shell=True, check=False)
        MityUtil.tabix(output_path)



"""
MITY Merge

Combines a nuclear VCF with MITY by adding MITY MT variants.
"""

import logging
import os.path
import sys
import gzip
import pysam
import pysam.bcftools
from mitylib.util import MityUtil


logger = logging.getLogger(__name__)


class Merge:
    """
    Combines a nuclear VCF with a MITY VCF
    """

    def __init__(
        self,
        debug,
        nuclear_vcf_path,
        mity_vcf_path,
        genome,
        output_dir=".",
        prefix=None,
        keep=False,
    ):
        self.debug = debug
        self.nuclear_vcf_path = nuclear_vcf_path
        self.mity_vcf_path = mity_vcf_path
        self.genome = genome

        self.bcftools_isec_path = ""
        self.bcftools_concat_path = ""
        self.merged_sorted_vcf_path = ""
        self.merged_unsorted_vcf_path = ""

        self.output_dir = output_dir
        self.prefix = prefix
        self.keep = keep

        self.run()

    def run(self):
        """
        Run mity merge.
        """

        if self.debug:
            logger.setLevel(logging.DEBUG)
            logger.debug("Entered debug mode.")
        else:
            logger.setLevel(logging.INFO)

        self.run_checks()
        self.set_strings()
        self.run_bcftools_isec()
        self.run_bcftools_concat()
        self.write_merged()
        MityUtil.gsort(
            self.merged_unsorted_vcf_path, self.merged_sorted_vcf_path, self.genome
        )
        self.remove_intermediate_files()

    def run_checks(self):
        """
        Check whether the nuclear and mity vcfs are compatible based on:
            - contig matching
        """

        mity_contig = MityUtil.vcf_get_mt_contig(self.mity_vcf_path)
        nuclear_contig = MityUtil.vcf_get_mt_contig(self.nuclear_vcf_path)

        if mity_contig != nuclear_contig:
            logger.error("The VCF files use mitochondrial contigs.")
            sys.exit()

    def set_strings(self):
        """
        Sets:
            - bcftools_isec_path
            - bcftools_concat_path
            - merged_unsorted_vcf_path
            - merged_sorted_vcf_path
        """
        if self.prefix is None:
            self.prefix = MityUtil.make_prefix(self.mity_vcf_path)

        self.bcftools_isec_path = os.path.join(self.output_dir, "0000.vcf")
        self.bcftools_concat_path = os.path.join(
            self.output_dir, self.prefix + ".bcftools.concat.vcf"
        )

        self.merged_unsorted_vcf_path = os.path.join(
            self.output_dir, self.prefix + ".mity.merge.unsorted.vcf"
        )
        self.merged_sorted_vcf_path = os.path.join(
            self.output_dir, self.prefix + ".mity.merge.vcf.gz"
        )

    def run_bcftools_isec(self):
        """
        Run bcftools isec
        """

        with open(self.bcftools_isec_path, "w", encoding="utf-8") as file:
            print(
                pysam.bcftools.isec(
                    "-p",
                    self.output_dir,
                    "-C",
                    self.nuclear_vcf_path,
                    self.mity_vcf_path,
                ),
                end="",
                file=file,
            )

    def run_bcftools_concat(self):
        """
        Run bcftools concat
        """

        with open(self.bcftools_concat_path, "w", encoding="utf-8") as file:
            print(
                pysam.bcftools.concat(self.bcftools_isec_path, self.mity_vcf_path),
                end="",
                file=file,
            )

    def merge_description(self, nuclear_description: str, mity_description: str):
        """
        Merge nuclear and mity header description.
        """
        return f"If CHR=MT OR CHR=chrM: {mity_description}, otherwise: {nuclear_description}"

    def get_header_line_nums(self):
        """
        Returns a dictionary with key value pairs of ID : current position.

        E.g.
        ##FORMAT=<ID=AD,Number=.,Type=Integer,Description="Allelic depths ...">
        ##FORMAT=<ID=DP,Number=1,Type=Integer,Description="Approximate rea...">
        ##FORMAT=<ID=GQ,Number=1,Type=Integer,Description="Genotype Quality">
        ##FORMAT=<ID=GT,Number=1,Type=String,Description="Genotype">

        ##INFO=<ID=AC,Number=A,Type=Integer,Description="Allele count in g...">
        ##INFO=<ID=AF,Number=A,Type=Float,Description="Allele Frequency, f...">
        ##INFO=<ID=AN,Number=1,Type=Integer,Description="Total number of a...">
        ##INFO=<ID=BaseQRankSum,Number=1,Type=Float,Description="Z-score f...">


        Gives the dictionary:
        dict = {
            FORMAT: {
                AD: 0,      # this represents the line starting from 0 bytes
                DP: 80,     # this line starts from 80 bytes
                GQ: 160,
                GT: ...
            },
            INFO: {
                AC: ...,
                AF: ...,
                AN: ...,
                BaseQRankSum: ...
            }
        }

        This is then used to file.seek later. Note that this is a workaround to
        reduce memory footprint (it only stores position), but also give
        O(n + m) performance instead of O(nm) if we simply loop over each file
        from the beginning each time.
        """

        line_nums = {"FORMAT": {}, "INFO": {}}

        with gzip.open(self.mity_vcf_path, "rt") as file:
            position = 0
            file.seek(0)

            while True:
                position = file.tell()

                line = file.readline()

                if not line:
                    break

                if not line.startswith("##"):
                    break

                if line.startswith("##FORMAT") or line.startswith("##INFO"):
                    section, field_id = self.get_header_line_info(line)
                    line_nums[section][field_id] = position

        return line_nums

    def get_header_line_info(self, line: str):
        """
        Get section (INFO/FORMAT) and ID of a header line.
        """
        if line.startswith("##FORMAT"):
            section = "FORMAT"
        else:
            section = "INFO"

        line = line.split(",")[0]
        field_id = line.split("=")[-1]

        return section, field_id

    def get_header_description(self, line: str):
        """
        Get header description.
        """
        if "Description=" not in line:
            return ""

        return line.split("Description=")[1].strip('>" ')

    def make_new_line(self, nuclear_line, mity_file_line_num):
        """
        Makes an updated line with merged description.
        """
        with gzip.open(self.mity_vcf_path, "rt") as file:
            file.seek(mity_file_line_num)
            mity_line = file.readline()

        mity_line = mity_line.strip("\n")
        nuclear_line = nuclear_line.strip("\n")

        nuclear_description = self.get_header_description(nuclear_line)
        mity_description = self.get_header_description(mity_line)

        new_description = self.merge_description(nuclear_description, mity_description)

        new_line = nuclear_line.split("Description=")[0]
        new_line += f'Description="{new_description}">\n'

        return new_line

    def write_merged(self):
        """
        Make a new file with updated headers and add variants.
        """

        header_dict = self.get_header_line_nums()

        with open(self.merged_unsorted_vcf_path, "w", encoding="utf-8") as merged_file:
            with open(self.bcftools_concat_path, "r") as concat_file:
                for line in concat_file:
                    if line.startswith("##FORMAT") or line.startswith("##INFO"):
                        section, field_id = self.get_header_line_info(line)
                        if field_id in header_dict[section]:
                            new_line = self.make_new_line(
                                line, header_dict[section][field_id]
                            )
                            merged_file.write(new_line)
                        else:
                            merged_file.write(line)
                    else:
                        merged_file.write(line)

    def remove_intermediate_files(self):
        """
        Removes extra files from bcftools isec:
            README.txt
            sites.txt

        Optionally removes:
            bcftools.isec file
            bcftools.concat file
        """
        os.remove(os.path.join(self.output_dir, "README.txt"))
        os.remove(os.path.join(self.output_dir, "sites.txt"))

        if not self.keep:
            os.remove(self.merged_unsorted_vcf_path)
            os.remove(self.bcftools_isec_path)
            os.remove(self.bcftools_concat_path)



"""Mitochondrial variant calling."""
import subprocess
import logging
import os.path
import sys
import urllib.request

import pysam

from mitylib.normalise import Normalise
from mitylib.util import MityUtil

logger = logging.getLogger(__name__)


class Call:
    """
    Mity call.
    """

    MIN_MQ = 30
    MIN_BQ = 24
    MIN_AF = 0.01
    MIN_AC = 4
    P_VAL = 0.002

    def __init__(
        self,
        debug,
        files,
        reference,
        genome=None,
        prefix=None,
        min_mq=MIN_MQ,
        min_bq=MIN_BQ,
        min_af=MIN_AF,
        min_ac=MIN_AC,
        p=P_VAL,
        normalise=True,
        output_dir=".",
        region=None,
        bam_list=False,
        keep=False,
    ):
        self.debug = debug
        self.files = files[0]
        self.reference = reference
        self.genome = genome
        self.prefix = prefix
        self.min_mq = min_mq
        self.min_bq = min_bq
        self.min_af = min_af
        self.min_ac = min_ac
        self.p = p
        self.normalise = normalise
        self.output_dir = output_dir
        self.region = region
        self.bam_list = bam_list
        self.keep = keep

        self.file_string = ""
        self.normalised_vcf_path = ""
        self.call_vcf_path = ""

        self.mity_cmd = ""
        self.sed_cmd = ""

        self.run()

    def run(self):
        """
        Run mity call.
        """

        if self.debug:
            logger.setLevel(logging.DEBUG)
            logger.debug("Entered debug mode.")
        else:
            logger.setLevel(logging.INFO)

        if (self.bam_list):
            self.get_files_from_list()
        self.run_checks()
        self.set_strings()
        self.set_region()
        self.set_mity_cmd()

        self.run_freebayes()

        if self.normalise:
            self.run_normalise()
        else:
            MityUtil.tabix(self.call_vcf_path)

    def run_normalise(self):
        """
        Run mity normalise.
        """
        logger.debug("Normalising and Filtering variants")

        try:
            Normalise(
                debug=self.debug,
                vcf=self.call_vcf_path,
                reference_fasta=self.reference,
                prefix=self.prefix,
                output_dir=self.output_dir,
                allsamples=False,
                p=self.p,
                genome=self.genome,
                keep=self.keep,
            )
        finally:
            if not self.keep:
                os.remove(self.call_vcf_path)

    def run_freebayes(self):
        """
        Run freebayes.
        """
        freebayes_call = (
            f"set -o pipefail && freebayes -f {self.reference} {self.file_string} "
            f"--min-mapping-quality {self.min_mq} "
            f"--min-base-quality {self.min_bq} "
            f"--min-alternate-fraction {self.min_af} "
            f"--min-alternate-count {self.min_ac} "
            f"--ploidy 2 "
            f"--region {self.region} "
            f"| sed 's/##source/##freebayesSource/' "
            f"| sed 's/##commandline/##freebayesCommandline/' "
            f"| {self.sed_cmd} | bgzip > {self.call_vcf_path}"
        )

        logger.info("Running FreeBayes in sensitive mode")
        logger.debug(freebayes_call)
        res = subprocess.run(
            freebayes_call,
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            executable="/bin/bash",
            check=False,
        )
        logger.debug("Freebayes result code: %s", res.returncode)

        if res.returncode != 0:
            logger.error("FreeBayes failed: %s", res.stderr)
            exit(1)

        if os.path.isfile(self.call_vcf_path):
            logger.debug("Finished running FreeBayes")

    def set_mity_cmd(self):
        """
        Creates the mity command for embedding into the vcf.
        """

        mity_cmd = (
            f'##mityCommandline="mity call --reference {self.reference} --prefix {self.prefix} '
            f"--min-mapping-quality {self.min_mq} --min-base-quality {self.min_bq} "
            f"--min-alternate-fraction {self.min_af} --min-alternate-count {self.min_ac} "
            f"--out-folder-path {self.output_dir} --region {self.region}"
        )

        if self.normalise:
            mity_cmd += f" --normalise --p {self.p}"

        mity_cmd += " " + " ".join(self.files)
        mity_cmd += '"'
        mity_cmd = mity_cmd.replace("/", "\\/")

        logger.debug(mity_cmd)

        # overwrite a redundant freebayes header line with the mity command line
        sed_cmd = f"sed 's/^##phasing=none/{mity_cmd}/g'"
        logger.debug(sed_cmd)

        self.mity_cmd = mity_cmd
        self.sed_cmd = sed_cmd

    def set_region(self):
        """
        Sets the region if not specified.
        """
        if self.region is None:
            self.region = self.bam_get_mt_contig(self.files[0], as_string=True)

    def set_strings(self):
        """
        Sets:
            prefix
            file_string
            call_output_file
            normalise_output_file
        """
        if self.prefix is None:
            self.prefix = self.make_prefix(self.files[0], self.prefix)

        self.file_string = " ".join(["-b " + _file for _file in reversed(self.files)])

        self.normalised_vcf_path = os.path.join(
            self.output_dir, self.prefix + ".mity.normalise.vcf.gz"
        )
        self.call_vcf_path = os.path.join(
            self.output_dir, self.prefix + ".mity.call.vcf.gz"
        )

    def run_checks(self):
        """
        Check for valid input.
        """

        if len(self.files) > 1 and self.prefix is None:
            raise ValueError(
                "If there is more than one bam/cram file, --prefix must be set"
            )

        self.check_missing_file(self.files, die=True)
        self.prefix = self.make_prefix(self.files[0], self.prefix)

        if not all(map(self.bam_has_rg, self.files)):
            logger.error("At least one BAM/CRAM file lacks an @RG header")
            exit(1)

        if self.normalise and self.genome is None:
            logger.error("A genome file should be supplied if mity call normalize=True")
            sys.exit(1)

    def bam_has_rg(self, bam):
        """
        Check whether a BAM or CRAM file contains a valid @RG header, which is critical for accurate variant calling with mity.

        Parameters:
            - bam (str): Path to a BAM or CRAM file.

        Returns:
            - bool: True if the file has a valid @RG header, False otherwise.
        """
        r = pysam.AlignmentFile(bam, "rb")
        return len(r.header["RG"]) > 0

    def bam_get_mt_contig(self, bam: str, as_string: bool = False):
        """
        Retrieve mitochondrial contig information from a BAM or CRAM file.

        Parameters:
            - bam (str): Path to a BAM or CRAM file.

        Keyword Arguments:
            - with_coordinates (bool): If True, the result includes the contig
            name with coordinates (e.g., 'chrM:1-16569'). If False, it returns a
            tuple containing the contig name (str) and its length (int).

        Returns:
            - If with_coordinates is False, a tuple (contig_name, contig_length).
            - If with_coordinates is True, a string with coordinates.
        """
        r = pysam.AlignmentFile(bam, "rb")
        chroms = [str(record.get("SN")) for record in r.header["SQ"]]
        mito_contig = {"MT", "chrM"}.intersection(chroms)
        assert len(mito_contig) == 1
        mito_contig = "".join(mito_contig)
        res = None
        for record in r.header["SQ"]:
            if mito_contig == record["SN"]:
                res = record["SN"], record["LN"]
        if res is not None and as_string:
            res = res[0] + ":1-" + str(res[1])
        return res

    def make_prefix(self, file_name: str, prefix: str = None):
        """
        Generate a prefix for Mity functions if a custom prefix is not provided,
        the function uses the filename without the file extension (.vcf, .bam, .cram, .bed).

        Parameters:
            file_name (str): The filename, including extensions (e.g., .vcf, .bam, .cram, .bed).
            prefix (str, optional): An optional custom prefix. If None, the function generates a prefix from the file name.

        Returns:
            str: The generated or custom prefix for the Mity function.
        """
        supported_extensions = {".vcf": ".vcf", ".bam": ".bam", ".cram": ".cram"}

        ext = os.path.splitext(file_name)[1]
        if ext in supported_extensions:
            return (
                prefix
                if prefix is not None
                else os.path.basename(file_name).replace(ext, "")
            )
        else:
            raise ValueError("Unsupported file type")
        
    def get_files_from_list(self, die: bool = True):
        """
        Get the list of BAM / CRAM files from the provided file
        """
        if len(self.files) > 1:
            raise ValueError("--bam-file-list Argument expects only 1 file to be provided.")
        with open(self.files[0], 'r') as f:
            self.files = f.read().splitlines()


    def check_missing_file(self, file_list: str, die: bool = True):
        """
        Check if input files exist.
        """
        missing_files = []
        for item in file_list:
            if item.lower().startswith('http'):
                try:
                    urllib.request.urlopen(item).getcode()
                except:
                    missing_files.append(item)
            elif not os.path.isfile(item):
                missing_files.append(item)
        if die and len(missing_files) > 0:
            raise ValueError("Missing these files: " + ",".join(missing_files))
        return missing_files



"""
Adds annotations and generates mity excel report from VCF file.
"""

import logging
import os.path
import subprocess
import pysam
import pandas
import yaml

# LOGGING
logger = logging.getLogger(__name__)


class Vep:
    """
    Provides methods to handle vepped files.
    """

    def __init__(self, vcf_obj):
        self.vcf_obj = vcf_obj
        self.vepped = self.check_vepped()

        if self.vepped:
            self.vep_keys = self.generate_vep_keys()
            self.vep_excel_headers = self.generate_vep_excel_headers()

    def get_vep_name(self, name):
        """
        Capitalises vep name, replaces _ with <space> and adds "VEP" to the end.
        """
        name = name.upper()
        name = name.replace("_", " ")
        name = name + " VEP"

        return name

    def generate_vep_excel_headers(self):
        """
        Generates vep excel headers based on the vep keys.
        """
        vep_excel_headers = ["HIGHEST IMPACT VEP"]
        for name in self.vep_keys:
            vep_excel_headers.append(self.get_vep_name(name))

        return vep_excel_headers

    def find_highest_impact(self, impacts):
        """
        Returns the highest impact in the list of consequences. Raises error if
        none of the consequences match.
        """
        if "HIGH" in impacts:
            return "HIGH"
        if "MODERATE" in impacts:
            return "MODERATE"
        if "MODIFIER" in impacts:
            return "MODIFIER"
        if "LOW" in impacts:
            return "LOW"

        logging.warning("Unknown SO term.")

        return ""

    def get_vep_values(self, variant):
        """
        Takes a string from VEP consequences/impacts in the form:
            impact value | impact value | ... |, |||, |||

        And peforms the following:
            - removes annotation "line" if the consequence has "stream" in it,
              e.g. "upstream"
            - concatenates remaining line fields with ";"

        Types:
            variant.info["CSQ"]: Tuple of form (a|b|c|..., a|b|c|..., a|b|c|...)

        Example:
            vep_keys = [ IMPACT, Consequence, field_1, field_2, field_3 ]

            list(variant.info["CSQ"] = [
                HIGH        | something_else    | a | b | c,
                LOW         | upstream_variant  | d | e | f,
                MODIFIER    | something_else    | g | h | i
            ]

            vep_dict = {
                HIGHEST IMPACT VEP  : HIGH,
                CONSEQUENCE VEP     : something_else;something_else,
                FIELD 1 VEP         : a;g,
                FIELD 2 VEP         : b;h,
                FIELD 3 VEP         : c;i
            }
        """

        vep_dict = {}
        for name in self.vep_excel_headers:
            vep_dict[name] = []

        impacts = []

        for line in list(variant.info["CSQ"]):
            split_line = line.split("|")

            # NOTE: rename IMPACT if the name changes to something else
            impact = split_line[self.vep_keys.index("IMPACT")]
            if "stream" in impact:
                continue

            impacts.append(impact)

            for i, value in enumerate(split_line):
                if value:
                    vep_dict[self.vep_excel_headers[i + 1]].append(value)

        vep_dict["HIGHEST IMPACT VEP"] = self.find_highest_impact(impacts)

        # NOTE: final string format/join for excel output can be changed here
        for key, value in vep_dict.items():
            if isinstance(value, list):
                vep_dict[key] = ";".join(value)

        return vep_dict

    def check_vepped(self):
        """
        Checks whether a file has been vepped by looking for VEP in the vcf
        header. The line should look something like this:
            ##VEP="v110" time="2023-10-04 00:20:30" cache="/net/isilonP...
        """
        if "VEP" in str(self.vcf_obj.header):
            return True
        return False

    def generate_vep_keys(self):
        """
        Example header INFO line:
            ##INFO=<ID=CSQ,Number=.,Type=String,Description=
            "Consequence annotations from Ensembl VEP. Format:
            Allele|Consequence|IMPACT|SYMBOL|Gene|Feature_type|Feature|...

        Returns a list of keys, i.e. ["Allele", "Consequence", ...]

        NOTE: This method is hard coded based on the format of the description.
        """
        for header in self.vcf_obj.header.records:
            if header.type == "INFO" and "CSQ" in str(header) and header.attrs:
                description = header.attrs[-2][1].strip('"')

                # change this line if the description text or format changes
                description = description.replace(
                    "Consequence annotations from Ensembl VEP. Format:", ""
                )

                keys = description.split("|")
                return keys

        return None


class SingleReport:
    """
    Handles generating a mity report for one VCF file.
    """

    def __init__(self, vcf_path, min_vaf, keep) -> None:
        self.min_vaf = min_vaf
        self.keep = keep

        self.vcf_path = vcf_path
        self.vcf_obj = pysam.VariantFile(vcf_path)

        self.annot_vcf_path = None
        self.annot_vcf_obj = None

        # excel and vcf headers
        with open("report-config.yaml", "r") as report_config_file:
            self.report_config = yaml.safe_load(report_config_file)

        self.excel_headers = self.report_config["excel_headers"]
        self.vcf_headers = self.report_config["vcf_headers"]

        self.excel_table = {}
        self.df = None

        self.vep = None

        self.run()

    def run(self):
        """
        Run SingleReport.
        """
        self.vcfanno_call()
        self.vep = Vep(self.annot_vcf_obj)
        self.make_table()

        if not self.keep:
            # remove vcfanno annotated vcf
            os.remove(self.annot_vcf_path)

    def get_df(self):
        """
        Return a pandas dataframe from the generated excel table.
        """
        all_excel_headers = self.excel_headers
        if self.vep.vepped:
            all_excel_headers += self.vep.vep_excel_headers
        return pandas.DataFrame(self.excel_table, columns=all_excel_headers)

    def make_hgvs(self, pos, ref, alt):
        """
        Creates HGVS syntax used in the HGVS column of the table/excel spreadsheet.
        """
        if len(alt) > 1 or len(ref) > 1:
            # this is an indel
            if len(ref) > len(alt):
                # this is a del
                delet = ref[1:]
                if len(delet) == 1:
                    hgvs_pos = int(pos) + 1
                elif len(delet) > 1:
                    hvgs_pos_start = int(pos) + 1
                    hvgs_pos_end = int(pos) + len(delet)
                    hgvs_pos = str(hvgs_pos_start) + "_" + str(hvgs_pos_end)
                hgvs = "m." + str(hgvs_pos) + "del"

            else:
                # this is an ins
                ins = alt[1:]
                if len(ins) == 1:
                    hgvs_pos = int(pos) + 1
                elif len(ins) > 1:
                    hvgs_pos_start = int(pos) + 1
                    hvgs_pos_end = int(pos) + len(ins)
                    hgvs_pos = str(hvgs_pos_start) + "_" + str(hvgs_pos_end)
                hgvs = "m." + str(hgvs_pos) + "ins"

        else:
            # this is a SNP
            hgvs = "m." + str(pos) + str(ref) + ">" + str(alt)
        return hgvs

    def vcfanno_call(self):
        """
        Calls vcfanno to annotate the output of mity normalise with the relevant
        annotations based on vcfanno-config.toml in mitylib/
        """
        logger.debug("Running vcfanno...")

        # annotated_file name
        annotated_file = self.vcf_path.replace(".vcf.gz", ".mity.annotated.vcf")

        # vcfanno call
        vcfanno_cmd = (
            f"vcfanno -p 4 vcfanno-config.toml {self.vcf_path} > {annotated_file}"
        )
        res = subprocess.run(
            vcfanno_cmd,
            shell=True,
            check=False,
            # capture the output of the vcfanno command since it tends to produce
            # long warning messages, output shown in --debug mode
            capture_output=True,
            text=True,
        )

        logger.debug("vcfanno output:")
        logger.debug(res.stdout)

        self.annot_vcf_path = annotated_file
        self.annot_vcf_obj = pysam.VariantFile(annotated_file)

    def make_info_string(self, variant):
        """
        Takes the INFO fields from the variant and recreates the format as it
        appears in a vcf file. i.e.

        FIELD=X;FIELD=Y;etc
        """
        info_field_array = []
        for key, value in variant.info.items():
            if isinstance(value, tuple):
                value = value[0]
            info_field_array.append(f"{key}={value}")
        info_field_string = ";".join(info_field_array)
        return info_field_string

    def make_format_string(self, sample):
        """
        Takes the FORMAT fields from the sample and recreates the format as it
        appears in a vcf file. i.e.

        FIELD=X;FIELD=Y;etc
        """
        format_field_array = []
        for value in sample.values():
            if isinstance(value, tuple):
                value = value[0]
            format_field_array.append(str(value))
        format_field_string = ":".join(format_field_array)
        return format_field_string

    def clean_string(self, s):
        """
        Removes the following characters from a string:
            "
            '
            ()

        Used to output text from annotation sources.
        """
        if isinstance(s, tuple):
            s = s[0]

        s = str(s)
        s = s.replace('"', "").replace("(", "").replace(")", "")

        return s

    def make_table(self):
        """
        Takes a vcfanno annotated vcf file and returns a formatted dictionary
        with relevant information.

        The vcf and excel header names are hardcoded in report-config.yaml.
        """

        for header in self.excel_headers:
            self.excel_table[header] = []

        # check if VCF file is vepped, adds relevant VEP headers

        if self.vep.vepped:
            for header in self.vep.vep_excel_headers:
                self.excel_table[header] = []

        num_samples = len(self.annot_vcf_obj.header.samples)

        for variant in self.annot_vcf_obj.fetch():
            cohort_count = 0
            info_string = self.make_info_string(variant)

            # samples
            for sample in variant.samples.values():
                # skip sample if the VAF is too low
                if float(sample["VAF"][0]) <= float(self.min_vaf):
                    continue

                cohort_count += 1

                self.excel_table["SAMPLE"].append(sample.name)
                self.excel_table["HGVS"].append(
                    self.make_hgvs(variant.pos, variant.ref, variant.alts[0])
                )

                self.excel_table["CHR"].append(variant.chrom)
                self.excel_table["POS"].append(variant.pos)
                self.excel_table["REF"].append(variant.ref)

                # NOTE: assumes only one ALT
                self.excel_table["ALT"].append(variant.alts[0])
                self.excel_table["QUAL"].append(variant.qual)
                self.excel_table["FILTER"].append(variant.filter.keys()[0])

                self.excel_table["INFO"].append(info_string)
                self.excel_table["FORMAT"].append(self.make_format_string(sample))

                # vcf_headers: info
                for vcf_header, excel_header in self.vcf_headers["info"].items():
                    if vcf_header in variant.info.keys():
                        self.excel_table[excel_header].append(
                            self.clean_string(variant.info[vcf_header])
                        )
                    else:
                        self.excel_table[excel_header].append(".")

                # vcf_headers: annotations
                for vcf_header, excel_header in self.vcf_headers["annotations"].items():
                    if vcf_header in variant.info.keys():
                        self.excel_table[excel_header].append(
                            self.clean_string(variant.info[vcf_header])
                        )
                    else:
                        self.excel_table[excel_header].append(".")

                # vcf_headers: format
                for vcf_header, excel_header in self.vcf_headers["format"].items():
                    if vcf_header in sample.keys():
                        self.excel_table[excel_header].append(
                            self.clean_string(sample[vcf_header])
                        )
                    else:
                        self.excel_table[excel_header].append(".")

                # calculate ALLELE FREQUENCY MITOMAP
                # TODO: currently this formula matches old mity behaviour, check
                # that this is still valid.
                if "Genbank_frequency_mitomap" in sample.keys():
                    allele_frequency_mitomap = round(
                        float(sample["GenBank_frequency_mitomap"]) / 32059, 3
                    )
                else:
                    allele_frequency_mitomap = "."

                self.excel_table["ALLELE FREQUENCY MITOMAP"].append(
                    allele_frequency_mitomap
                )

            # fill in cohort count
            cohort_frequency = float(cohort_count / num_samples)
            for _ in range(cohort_count):
                self.excel_table["COHORT COUNT"].append(cohort_count)
                self.excel_table["COHORT FREQUENCY"].append(cohort_frequency)

            # fill in VEP impacts if vepped
            if self.vep.vepped:
                self.add_vep_impacts(variant, cohort_count)

    def add_vep_impacts(self, variant, cohort_count):
        """
        Adds vep impacts for a variant.
        """
        variant_impacts = self.vep.get_vep_values(variant)
        for _ in range(cohort_count):
            for vep_header in self.vep.vep_excel_headers:
                self.excel_table[vep_header].append(variant_impacts[vep_header])


class Report:
    """
    Runs mity report.
    """

    def __init__(
        self, debug, vcfs, prefix=None, min_vaf=0.0, output_dir=".", keep=False
    ):
        self.debug = debug
        self.vcfs = vcfs[0]
        self.prefix = prefix
        self.min_vaf = min_vaf
        self.output_dir = output_dir
        self.keep = keep

        self.run()

    def run(self):
        """
        Runs mity report.
        """
        if self.debug:
            logger.setLevel(logging.DEBUG)
            logger.debug("Entered debug mode.")
        else:
            logger.setLevel(logging.INFO)

        if not os.path.exists(self.output_dir):
            os.makedirs(self.output_dir)

        xlsx_name = os.path.join(
            self.output_dir, self.prefix + ".annotated_variants.xlsx"
        )
        with pandas.ExcelWriter(xlsx_name, engine="xlsxwriter") as writer:
            if isinstance(self.vcfs, str):
                self.vcfs = [self.vcfs]
            for vcf in self.vcfs:
                single_report = SingleReport(vcf, self.min_vaf, self.keep)
                df = single_report.get_df()

                sheet_name = vcf.replace(".vcf.gz", "").split("/")[-1]
                df.to_excel(writer, sheet_name=sheet_name, index=False)
